package com.siemag.test.remote;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import javax.ejb.Singleton;
import javax.ejb.Stateful;
import javax.ejb.Stateless;
import javax.jms.JMSException;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

public abstract class ServiceLocatorAbstract {


  protected Context                                 context                      = null;


  /**
   * Cache for <Global JNDI Name> ,<Concrete Module>, for doing a lookup without a module name
   */
  protected final ConcurrentHashMap<String, Module> moduleCache                  = new ConcurrentHashMap<String, Module>();

  public ServiceLocatorAbstract() {
    try {
      init();
    } catch (NamingException e) {
      e.printStackTrace();
    }
  }

  /**
   * Encapsulate the name of the jar that contains services.
   * @author vuru
   *
   */
  public interface Module {
    /**
     * Name of the Jar, without the suffix .jar, which contains the service implementation
     * @return
     */
    public String getModuleName();
  }

  public interface JNDIName {

    /**
     * @param module The correspondent module object which contains the bean implementation.
     * @return Set modulename and return the same instance.
     */
    public JNDIName withModule(Module module);

    /**
     * @param beanName The deployed name of the Service.
     * @return Set JNDIName and return the same instance.
     */
    public JNDIName withBeanName(String beanName);

    /**
     * @param interfaze The remote or Local interface of the Service.
     * @return Set JNDIName and return the same instance.
     */
    public JNDIName withBusinessInterface(Class<?> interfaze);

    /**
     * Compute the serve or client side lookupstring.
     * @return
     */
    public String getLookupString();

    /**
     * Return a new instance of me.
     * @return
     */
    public JNDIName cloneMe();

    /**
     * A lookup String for the Entitymanager
     * @return
     */
    public String getLookupStringEntitymanager();

    /**
     * A lookup String for the TransactionFactory
     * @return
     */
    public String getLookupStringTopicConnectionFactory();

    /**
     * If prefix needed for lookup a jms que or topic
     * @return
     */
    public String getLookupStringPrefixJMSTopic();

  }

  public abstract class JNDINameAbstract implements JNDIName {
    public final static String SEPARATOR = "/";
    public final static String APPNAME   = "reconnectiontest";

    private String             beanName  = "";
    private Module             module;
    private Class<?>           businessInterface;

    @Override
    public JNDIName withModule(Module module){
      this.module = module;
      return this;
    }

    @Override
    public JNDIName withBeanName(String beanName){
      this.beanName = beanName;
      return this;
    }

    @Override
    public JNDIName withBusinessInterface(Class<?> interfaze){
      this.businessInterface = interfaze;
      return this;
    }

    @Override
    public String getLookupString(){
      StringBuilder builder = new StringBuilder();
      builder.append(getGlobalPrefix())
          .append(APPNAME).append(SEPARATOR);
      if (module != null) {
        builder.append(module.getModuleName()).append(SEPARATOR);
      } else {
        builder.append("").append(SEPARATOR);
      }

      builder.append(beanName);
      if (businessInterface != null) {
        builder.append("!").append(businessInterface.getName());
      }
      return builder.toString();
    }

    public String getBeanName(){
      return beanName;
    }

    public Module getModule(){
      return module;
    }

    public Class<?> getInterface(){
      return businessInterface;
    }

    /**
     * The first letter of the lookup String "java:global" for serverside for example.
     * @return
     */
    public abstract String getGlobalPrefix();

  }

  /**
   * Static class with the name of the modules wich contains EJB services
   * @author vuru
   *
   */
  public static class Modules {

    private static List<Module> modules;



    /**
     * DAO Servcices.
     */
    public static final Module  RECON_SERVICE      = new Module() {
                                                         @Override
                                                         public String getModuleName(){
                                                           return "reconnectiontest-ejb";
                                                         }
                                                       };



    /**
     * List of all registered modules
     */
    public static List<Module> values(){
      if (modules == null) {
        modules = new ArrayList<Module>();
        Field[] fields = Modules.class.getFields();
        for (Field field : fields) {
          if (field.getType().isAssignableFrom(Module.class)) {
            try {
              // All the module fields are static so we can use field.get(null)
              modules.add(Module.class.cast(field.get(null)));
            } catch (Exception e) {
              e.printStackTrace();
            }
          }
        }
      }
      return modules;
    }

    /**
     * The Module object for a given ModuleName
     */
    public static Module getModuleForName(String aModuleName){
      for (Module module : values()) {
        if (module.getModuleName().equals(aModuleName)) {
          return module;
        }
      }
      return null;
    }


  }


  private final TopicConnectionFactory getTopicConnectionFactory() throws NamingException{
    return (TopicConnectionFactory)context.lookup(getJNDINameProvider().getLookupStringTopicConnectionFactory());
  }




  public final Topic getTopic(String aName) throws NamingException{
    return (Topic)this.context.lookup(getJNDINameProvider().getLookupStringPrefixJMSTopic() + aName);
  }

  /**
   * Computes the deployed name of a bean.
   * @param beanClass
   * @return
   */
  public final String computeBeanName(Class<?> beanClass){
    Stateless stateless = (Stateless)beanClass.getAnnotation(Stateless.class);
    if (stateless != null && isNotEmpty(stateless.name())) {
      return stateless.name();
    }
    Stateful stateful = (Stateful)beanClass.getAnnotation(Stateful.class);
    if (stateful != null && isNotEmpty(stateful.name())) {
      return stateful.name();
    }
    Singleton singleton = (Singleton)beanClass.getAnnotation(Singleton.class);
    if (singleton != null && isNotEmpty(singleton.name())) {
      return singleton.name();
    }
    return beanClass.getSimpleName();
  }

  private boolean isNotEmpty(String name){
    return (name != null && !name.isEmpty());
  }

  /**
   * Creates a new connection to JMS MOM.
   * @return TopicConnection - connection
   * @throws JMSException
   * @throws NamingException
   */
  public final TopicConnection getTopicConnection() throws JMSException, NamingException{
    Hashtable<?, ?> environment = context.getEnvironment();
    String userName = (String)environment.get(Context.SECURITY_PRINCIPAL);
    String passwd = (String)environment.get(Context.SECURITY_CREDENTIALS);
    TopicConnectionFactory topicConnectionFactory = getTopicConnectionFactory();
    TopicConnection connection;
    if (userName == null || passwd == null) {
      connection = topicConnectionFactory.createTopicConnection();
    } else {
      connection = topicConnectionFactory.createTopicConnection(userName, passwd);
    }
    return connection;
  }

  public void init(Hashtable<Object, String> aProps) throws NamingException{
    this.context = new InitialContext(aProps);
  }

  /**
   * Get the Service for the given interface.The name of the service interface and the name of the service must be the same.
   * @param aBeanInterface Remote Interface
   * @return
   */
  public <T> T getService(Class<? extends T> aBeanInterface){
    return getService(aBeanInterface, aBeanInterface.getSimpleName());
  }

  /**
   * Initilaize the Serivcelocator. Especially the Initialcontext. For server or client lookup.
   * @throws NamingException
   */
  public abstract void init() throws NamingException;

  /**
   *
   * @param aBeanInterface Remote Interface
   * @param aModule Name of the Jar, without the suffix .jar, which contains the service implementation
   * @param aBeanName Name of the service.
   * @return
   */
  public <T> T getService(Class<? extends T> aBeanInterface, Module aModule, String aBeanName){
    //    String lookupString = APP_NAME + "/" + aBeanName + "/local-" + aBeanInterface.getName();
    JNDIName globalJNDIName = getJNDINameProvider().withBeanName(aBeanName)
        .withModule(aModule)
        .withBusinessInterface(aBeanInterface);
    //    ClassLoader classLoader = MaterialFlowControllerLocal.class.getClassLoader();
    //    ClassLoader contextClassLoader = Thread.currentThread().getContextClassLoader();
    //    Thread.currentThread().setContextClassLoader(classLoader);
    Object service = null;
    T locatedService = null;

    if (aModule != null) {
      try {
        service = context.lookup(globalJNDIName.getLookupString());
      } catch (NamingException e) {
        e.printStackTrace();
        return null;
      }
    } else {
      if (moduleCache.containsKey(globalJNDIName.getLookupString())) {
        return getService(aBeanInterface, moduleCache.get(globalJNDIName.getLookupString()), aBeanName);
      }
      for (Module module : Modules.values()) {
        globalJNDIName = globalJNDIName.withModule(module);
        System.out.println("Do lookup for "+globalJNDIName.getLookupString());
        try {
          service = context.lookup(globalJNDIName.getLookupString());
          moduleCache.put(globalJNDIName.cloneMe().withModule(null).getLookupString(), module);
          break;
        } catch (NamingException e) {
          // Try with next module
        }
      }
      if (service == null) {
        System.out.println("Lookup for service " + aBeanName + " failed ");
      }
    }
    try {
      //      debugLookupString(globalJNDIName, locatedService, aBeanInterface);
      locatedService = aBeanInterface.cast(service);
    } catch (ClassCastException e) {
      System.err.println("ClassCastException for servicename " + aBeanName + " with beaninterface " + aBeanInterface.getName());
    }
    //    finally{
    //      Thread.currentThread().setContextClassLoader(contextClassLoader);
    //    }

    return locatedService;
  }


  /**
   *
   * @param aBeanInterface aBeanInterface Remote Interface
   * @param aModule Name of the Jar, without the suffix .jar, which contains the service implementation
   * @param aBean Name of the service.
   * @return
   */
  public <T> T getService(Class<? extends T> aBeanInterface, Module aModule, Class<?> aBean){
    return getService(aBeanInterface, aModule, computeBeanName(aBean));
  }

  /**
   *
   * @param aBeanInterface Remote Interface
   * @param aBeanName Name of the service.
   * @return
   */
  public <T> T getService(Class<? extends T> aBeanInterface, String aBeanName){
    return getService(aBeanInterface, null, aBeanName);
  }

  /**
   *
   * @param aBeanInterface Remote Interface
   * @param aBean Class of the service
   * @return
   */
  public <T> T getService(Class<? extends T> aBeanInterface, Class<?> aBean){
    return getService(aBeanInterface, computeBeanName(aBean));
  }

  /**
   * The Client or server side JNDIName implementation
   * @return
   */
  public abstract JNDIName getJNDINameProvider();
}
